/**
 * @file Matrix_Power_Of_Hip_Ref_Async.cpp
 * @author Trasgo Group
 * @brief MatrixPow: Asynchronous native HIP version
 * @version 3.0
 * @date 2021-07-31
 *
 * @copyright This software is provided to enhance knowledge and encourage progress in the scientific
 * community. It should be used only for research and educational purposes. Any reproduction
 * or use for commercial purpose, public redistribution, in source or binary forms, with or
 * without modifications, is NOT ALLOWED without the previous authorization of the copyright
 * holder. The origin of this software must not be misrepresented; you must not claim that you
 * wrote the original software. If you use this software for any purpose (e.g. publication),
 * a reference to the software package and the authors must be included.
 *
 * @copyright THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * @copyright Copyright (c) 2007-2020, Trasgo Group, Universidad de Valladolid.
 * All rights reserved.
 *
 * @copyright More information on http://trasgo.infor.uva.es/
 */

// System includes
#include <assert.h>
#include <math.h>
#include <omp.h>
#include <stdio.h>

#include "hip/hip_runtime.h"

#ifdef _PROFILING_ENABLED_
#include <roctx.h>
#endif //_PROFILING_ENABLED_

#define SEED    6834723
#define EPSILON 0.0001

double main_clock;
double exec_clock;

#ifdef DEBUG
#define HIP_CHECK()                                                       \
	{                                                                     \
		hipError_t error;                                                 \
		if ((error = hipGetLastError()) != hipSuccess) {                  \
			fprintf(stdout, "%s::%d ERROR: %s: %s\n", __FILE__, __LINE__, \
					hipGetErrorName(error), hipGetErrorString(error));    \
			exit(EXIT_FAILURE);                                           \
		}                                                                 \
	}
#else
#define HIP_CHECK()
#endif

/*
 * function name: gpu_square_matrix_mult
 *
 * source: https://github.com/lzhengchun/matrix-cuda/blob/master/matrix_cuda.cu
 */
__global__ void matrixMulHIP(float *d_c, float *d_a, float *d_b, int n) {
	__shared__ float tile_a[BLOCKSIZE][BLOCKSIZE];
	__shared__ float tile_b[BLOCKSIZE][BLOCKSIZE];

	int   row = blockIdx.y * BLOCKSIZE + threadIdx.y;
	int   col = blockIdx.x * BLOCKSIZE + threadIdx.x;
	float tmp = 0.0;
	int   idx;

	for (int sub = 0; sub < gridDim.x; ++sub) {
		idx = row * n + sub * BLOCKSIZE + threadIdx.x;
		if (idx >= n * n) {
			// n may not divisible by BLOCKSIZE
			tile_a[threadIdx.y][threadIdx.x] = 0;
		} else {
			tile_a[threadIdx.y][threadIdx.x] = d_a[idx];
		}

		idx = (sub * BLOCKSIZE + threadIdx.y) * n + col;
		if (idx >= n * n) {
			tile_b[threadIdx.y][threadIdx.x] = 0;
		} else {
			tile_b[threadIdx.y][threadIdx.x] = d_b[idx];
		}
		__syncthreads();

		for (int k = 0; k < BLOCKSIZE; ++k) {
			tmp += tile_a[threadIdx.y][k] * tile_b[k][threadIdx.x];
		}
		__syncthreads();
	}
	if (row < n && col < n) {
		d_c[row * n + col] = tmp;
	}
}

typedef struct Init_Tiles_Args {
	float *matrix_a, *matrix_b, *matrix_c;
	int    rows, columns;
} Init_Tiles_Args_t;

void Init_Tiles(void *args) {
	Init_Tiles_Args_t *tmp      = (Init_Tiles_Args_t *)args;
	float             *matrix_a = tmp->matrix_a;
	float             *matrix_b = tmp->matrix_b;
	float             *matrix_c = tmp->matrix_c;
	int                rows     = tmp->rows;
	int                columns  = tmp->columns;

	srand(SEED);
	for (int j = 0; j < columns; j++) {
		float col_sum_a = 0;
		for (int i = 0; i < rows; i++) {
			// generate random floats in a way matrixes don't turn into NaN
			float min    = -(1 - col_sum_a) + EPSILON;
			float max    = 1 - col_sum_a - EPSILON;
			float random = ((float)rand()) / (float)RAND_MAX;
			float range  = max - min;
			float value  = (random * range) + min;

			matrix_a[i * columns + j] = value;
			matrix_b[i * columns + j] = value;
			matrix_c[i * columns + j] = 0;
			col_sum_a += fabsf(value);
		}
	}
}

typedef struct Host_Compute_Args {
	double *p_sum;
	double *p_res;
	float  *matrix;
	float  *matrix_res;
	int     rows;
	int     columns;
} Host_Compute_Args_t;

void Host_Compute(void *args) {
	#ifdef _PROFILING_ENABLED_
	roctxRangePush("Host task");
	#endif //_PROFILING_ENABLED_

	Host_Compute_Args_t *tmp = (Host_Compute_Args_t *)args;

	static int ITER       = 0;
	double    *p_sum      = tmp->p_sum;
	double    *p_res      = tmp->p_res;
	float     *matrix     = tmp->matrix;
	float     *matrix_res = tmp->matrix_res;
	int        rows       = tmp->rows;
	int        columns    = tmp->columns;

	double minimum = matrix[0 * columns + 0];
	double maximum = matrix[0 * columns + 0];

	for (int j = 0; j < rows; j++) {
		for (int k = 0; k < columns; k++) {
			if (minimum > matrix[j * columns + k]) {
				minimum = matrix[j * columns + k];
			}
			if (maximum < matrix[j * columns + k]) {
				maximum = matrix[j * columns + k];
			}
		}
	}

	for (int j = 0; j < rows; j++) {
		for (int k = 0; k < columns; k++) {
			matrix[j * columns + k] = matrix[j * columns + k] - minimum;
			matrix[j * columns + k] = matrix[j * columns + k] / maximum;
		}
	}

	p_sum[ITER] = 0;
	for (int j = 0; j < rows; j++) {
		for (int k = 0; k < columns; k++) {
			p_sum[ITER] += pow(matrix[j * columns + k], 2);
		}
	}
	p_res[ITER] = sqrt(p_sum[ITER]);

	ITER++;

	for (int j = 0; j < rows; j++) {
		for (int k = 0; k < columns; k++) {
			matrix_res[j * columns + k] = matrix[j * columns + k] / p_res[ITER];
		}
	}
	#ifdef _PROFILING_ENABLED_
	roctxRangePop();
	#endif //_PROFILING_ENABLED_
}

int main(int argc, char **argv) {
	main_clock = omp_get_wtime();

	// Simpler interface for square matrices (only one parameter)
	if (argc != 4) {
		fprintf(stderr, "Usage: %s <matrix_size> <n_power> <device>\n\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	int SIZE        = atoi(argv[1]);
	int POWER       = atoi(argv[2]);
	int GPU         = atoi(argv[3]);
	int MATRIX_SIZE = sizeof(float) * SIZE * SIZE;

	// Extra information for collecting results
	hipDeviceProp_t hip_dev_prop;
	hipGetDeviceProperties(&hip_dev_prop, GPU);
	#ifdef _CTRL_EXAMPLES_EXP_MODE_
	printf("%s, ", hip_dev_prop.name);
	#else // _CTRL_EXAMPLES_EXP_MODE_
	printf("\n ----------------------- ARGS ----------------------- \n");
	printf("\n SIZE: %d", SIZE);
	printf("\n N_ITER: %d", POWER);
	printf("\n DEVICE: %s", hip_dev_prop.name);
	printf("\n POLICY ASYNC");
	printf("\n\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_EXP_MODE_
	fflush(stdout);

	double *p_res = (double *)malloc(POWER * sizeof(double));
	double *p_sum = (double *)malloc(POWER * sizeof(double));

	hipSetDevice(GPU);
	HIP_CHECK();

	dim3 dimsA;
	dim3 dimsB;

	dimsA.x = dimsA.y = dimsB.x = dimsB.y = SIZE;

	dim3 threads(BLOCKSIZE, BLOCKSIZE);
	dim3 grid((dimsB.x + BLOCKSIZE - 1) / threads.x,
			  (dimsA.y + BLOCKSIZE - 1) / threads.y);

	/*Variables*/
	float *A, *B, *C;
	hipHostMalloc((void **)&A, MATRIX_SIZE);
	hipHostMalloc((void **)&B, MATRIX_SIZE);
	hipHostMalloc((void **)&C, MATRIX_SIZE);
	HIP_CHECK();
	float *matrix_tmp = (float *)malloc(MATRIX_SIZE);

	/*Device variables*/
	float *d_A, *d_B, *d_C;
	hipMalloc(&d_A, MATRIX_SIZE);
	hipMalloc(&d_B, MATRIX_SIZE);
	hipMalloc(&d_C, MATRIX_SIZE);
	HIP_CHECK();

	hipStream_t stream_DtH, stream_kernel, stream_host;

	hipStreamCreate(&stream_DtH);
	hipStreamCreate(&stream_kernel);
	hipStreamCreate(&stream_host);
	HIP_CHECK();

	hipEvent_t event_kernel, event_DtH_C, event_DtH_B, event_host_C, event_host_B;

	hipEventCreateWithFlags(&event_kernel, hipEventDisableTiming);
	hipEventCreateWithFlags(&event_DtH_C, hipEventDisableTiming);
	hipEventCreateWithFlags(&event_DtH_B, hipEventDisableTiming);
	hipEventCreateWithFlags(&event_host_C, hipEventDisableTiming);
	hipEventCreateWithFlags(&event_host_B, hipEventDisableTiming);

	HIP_CHECK();

	/*Initialize variables*/
	Init_Tiles_Args_t init_args = (Init_Tiles_Args_t){.matrix_a = A,
													  .matrix_b = B,
													  .matrix_c = C,
													  .rows     = SIZE,
													  .columns  = SIZE};
	hipLaunchHostFunc(stream_host, Init_Tiles, &init_args);

	Host_Compute_Args_t host_args_c =
		(Host_Compute_Args_t){.p_sum      = p_sum,
							  .p_res      = p_res,
							  .matrix     = C,
							  .matrix_res = matrix_tmp,
							  .rows       = SIZE,
							  .columns    = SIZE};
	Host_Compute_Args_t host_args_b =
		(Host_Compute_Args_t){.p_sum      = p_sum,
							  .p_res      = p_res,
							  .matrix     = B,
							  .matrix_res = matrix_tmp,
							  .rows       = SIZE,
							  .columns    = SIZE};

	/* Start timer */
	hipDeviceSynchronize();
	exec_clock = omp_get_wtime();

	/*Copy to the device*/
	hipMemcpyAsync(d_A, A, MATRIX_SIZE, hipMemcpyHostToDevice, stream_kernel);
	HIP_CHECK();

	hipMemcpyAsync(d_B, B, MATRIX_SIZE, hipMemcpyHostToDevice, stream_kernel);
	HIP_CHECK();

	hipMemcpyAsync(d_C, C, MATRIX_SIZE, hipMemcpyHostToDevice, stream_kernel);
	HIP_CHECK();

	/* Wait for data in the device to start first kernel computation */

	for (int i = 0; i < POWER; i++) {
		if ((i % 2) == 0) {
			hipStreamWaitEvent(stream_kernel, event_DtH_C, 0); // Wait for previous C transfer to finish
			matrixMulHIP<<<grid, threads, 0, stream_kernel>>>(d_C, d_A, d_B, SIZE);
			hipEventRecord(event_kernel, stream_kernel); // kernel C = A x B record
			HIP_CHECK();

			hipStreamWaitEvent(stream_DtH, event_kernel, 0); // Wait for kernel C = A x B
			hipStreamWaitEvent(stream_DtH, event_host_C, 0); // Wait for previous host function using C
			hipMemcpyAsync(C, d_C, MATRIX_SIZE, hipMemcpyDeviceToHost, stream_DtH);
			hipEventRecord(event_DtH_C, stream_DtH); // transfer Device to Host of C
			HIP_CHECK();

			hipStreamWaitEvent(stream_host, event_DtH_C, 0); // Wait for transfer of C to host
			hipLaunchHostFunc(stream_host, Host_Compute, &host_args_c);
			hipEventRecord(event_host_C, stream_host); // C is being used in the host
			HIP_CHECK();
		} else {
			hipStreamWaitEvent(stream_kernel, event_DtH_B, 0); // Wait for previous B transfer to finish
			matrixMulHIP<<<grid, threads, 0, stream_kernel>>>(d_B, d_A, d_C, SIZE);
			hipEventRecord(event_kernel, stream_kernel); // kernel B = A x C record
			HIP_CHECK();

			hipStreamWaitEvent(stream_DtH, event_kernel, 0); // Wait for kernel B = A x C
			hipStreamWaitEvent(stream_DtH, event_host_B, 0); // Wait for previous host function using B
			hipMemcpyAsync(B, d_B, MATRIX_SIZE, hipMemcpyDeviceToHost, stream_DtH);
			hipEventRecord(event_DtH_B, stream_DtH); // transfer Device to Host of B
			HIP_CHECK();

			hipStreamWaitEvent(stream_host, event_DtH_B, 0); // Wait for transfer of B to host
			hipLaunchHostFunc(stream_host, Host_Compute, &host_args_b);
			hipEventRecord(event_host_B, stream_host); // B is being used in the host
			HIP_CHECK();
		}
	}

	/* Stop timer */
	hipDeviceSynchronize();
	exec_clock = omp_get_wtime() - exec_clock;

	hipEventDestroy(event_kernel);
	hipEventDestroy(event_DtH_C);
	hipEventDestroy(event_DtH_B);
	hipEventDestroy(event_host_C);
	hipEventDestroy(event_host_B);
	hipStreamDestroy(stream_DtH);
	hipStreamDestroy(stream_kernel);
	hipStreamDestroy(stream_host);
	hipFree(d_A);
	hipFree(d_B);
	hipFree(d_C);
	HIP_CHECK();

	/* PRINT RESULTS */
	#ifdef _CTRL_EXAMPLES_EXP_MODE_
	printf("%lf, %lf, ", p_sum[POWER - 1], p_res[POWER - 1]);
	#else // _CTRL_EXAMPLES_EXP_MODE_
	printf("\n ----------------------- NORM ----------------------- \n\n");
	for (int i = 0; i < POWER; i++) {
		printf(" iter: %d, sum: %lf, res: %lf\n", i + 1, p_sum[i], p_res[i]);
	}
	printf("\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_EXP_MODE_
	fflush(stdout);

	hipHostFree(A);
	hipHostFree(B);
	hipHostFree(C);
	free(p_sum);
	free(p_res);
	free(matrix_tmp);

	main_clock = omp_get_wtime() - main_clock;

	#ifdef _CTRL_EXAMPLES_EXP_MODE_
	printf("%lf, %lf\n", main_clock, exec_clock);
	#else // _CTRL_EXAMPLES_EXP_MODE_
	printf("\n ---------------------- TIMERS ---------------------- \n");
	printf("Clock main: %lf\n", main_clock);
	printf("Clock exec: %lf\n", exec_clock);
	printf("\n\n ---------------------------------------------------- \n");
	#endif // _CTRL_EXAMPLES_EXP_MODE_

	return EXIT_SUCCESS;
}
