#include <assert.h>
#include <math.h>
#include <omp.h>
#include <stdio.h>
#include <stdlib.h>

#include "Ctrl.h"

#define policy	(Ctrl_Policy)1
#define DEVICE	0

#define SEED    6834723
#define EPSILON 0.0001

int SIZE;

Ctrl_NewType(float);

double main_clock;
double exec_clock;

CTRL_KERNEL_CHAR(Mult, MANUAL, LOCAL_SIZE, LOCAL_SIZE);

#define mult_params	3, OUT, HitTile_float, matrix_result, IN, HitTile_float, matrix_a, IN, HitTile_float, matrix_b

#define init_params	3, OUT, HitTile_float, matrix_a, OUT, HitTile_float, matrix_b, OUT, HitTile_float, matrix_c

#define host_task_params	5, INVAL, int, ITER, INVAL, double *, p_sum, INVAL, double *, p_res, IN, HitTile_float, matrix, INVAL, HitTile_float, matrix_res

CTRL_KERNEL(Mult, OPENCLGPU, DEFAULT, CTRL_KPARAMS( mult_params ), {
	__local float tile_a[LOCAL_SIZE][LOCAL_SIZE];
	__local float tile_b[LOCAL_SIZE][LOCAL_SIZE];

	int n_groups_x = get_num_groups(0);

	int local_size = get_local_size(0);

	int local_x = get_local_id(0);
	int local_y = get_local_id(1);

	float tmp = 0.0;
	int   idx;

	for (int sub = 0; sub < n_groups_x; ++sub) {
		idx = thread_id_x * hit_tileDimCard(matrix_a, 0) + sub * local_size + local_x;
		tile_a[local_y][local_x] = hit(matrix_a, idx);
		idx = (sub * local_size + local_y) * hit_tileDimCard(matrix_a, 0) + thread_id_y;
		tile_b[local_y][local_x] = hit(matrix_b, idx);
		barrier(CLK_LOCAL_MEM_FENCE);

		for (int k = 0; k < local_size; ++k) {
			tmp += tile_a[local_y][k] * tile_b[k][local_x];
		}
		barrier(CLK_LOCAL_MEM_FENCE);
	}
	hit(matrix_result, thread_id_x, thread_id_y) = tmp;
});

CTRL_HOST_TASK(Init_Tiles, CTRL_HPARAMS( init_params ) ) {
	srand(SEED);
	for (int j = 0; j < SIZE; j++) {
		float col_sum_a = 0;
		for (int i = 0; i < SIZE; i++) {
			float min = -(1 - col_sum_a) + EPSILON;
			float max = 1 - col_sum_a - EPSILON;
			float random = ((float)rand()) / (float)RAND_MAX;
			float range  = max - min;
			float value = (random * range) + min;
			hit(matrix_a, i, j) = value;
			hit(matrix_b, i, j) = value;
			hit(matrix_c, i, j) = 0;
			col_sum_a += fabsf(value);
		}
	}
}

CTRL_HOST_TASK(Host_Compute, CTRL_HPARAMS( host_task_params ) ) {
	double minimum = hit(matrix, 0);
	double maximum = hit(matrix, 0);

	for (int i = 0; i < SIZE*SIZE; i++) {
		if (minimum > hit(matrix, i)) {
			minimum = hit(matrix, i);
		}
		if (maximum < hit(matrix, i)) {
			maximum = hit(matrix, i);
		}
	}

	for (int i = 0; i < SIZE*SIZE; i++) {
		hit(matrix, i) -= minimum;
		hit(matrix, i) /= maximum;
	}

	p_sum[ITER] = 0;
	for (int i = 0; i < SIZE*SIZE; i++) {
		p_sum[ITER] += pow(hit(matrix, i), 2);
	}
	p_res[ITER] = sqrt(p_sum[ITER]);

	for (int i = 0; i < SIZE*SIZE; i++) {
		hit(matrix_res, i) = hit(matrix, i) / p_res[ITER];
	}
}

CTRL_KERNEL_PROTO(Mult, 1, OPENCLGPU, DEFAULT, mult_params );
CTRL_HOST_TASK_PROTO(Init_Tiles, init_params );
CTRL_HOST_TASK_PROTO(Host_Compute, host_task_params );

int main(int argc, char *argv[]) {
	main_clock = omp_get_wtime();

	if (argc != 4) {
		fprintf(stderr, "\nUsage: %s <size> <n_iter> <platform>\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	SIZE     = atoi(argv[1]);
	int         N_ITER   = atoi(argv[2]);
	int         PLATFORM = atoi(argv[4]);


	double *p_res = (double *)malloc(N_ITER * sizeof(double));
	double *p_sum = (double *)malloc(N_ITER * sizeof(double));

	Ctrl_Thread threads;
	Ctrl_ThreadInit(threads, SIZE, SIZE);

	__ctrl_block__(1, 0)
	{
		PCtrl ctrl = Ctrl_Create(CTRL_TYPE_OPENCL_GPU, policy, DEVICE, PLATFORM);

		HitShape shape = hitShapeSize(SIZE, SIZE);
		HitTile_float matrix_a = Ctrl_DomainAlloc(ctrl, float, shape);
		HitTile_float matrix_b = Ctrl_DomainAlloc(ctrl, float, shape);
		HitTile_float matrix_c = Ctrl_DomainAlloc(ctrl, float, shape);

		HitTile_float matrix_tmp = hitTile(float, shape);

		Ctrl_HostTask(ctrl, Init_Tiles, matrix_a, matrix_b, matrix_c);

		Ctrl_GlobalSync(ctrl);
		exec_clock = omp_get_wtime();

		for (int i = 0; i < N_ITER; i++) {
			HitTile_float matrix1, matrix2;
       			matrix1 = (i%2==0) ? matrix_c : matrix_b;
       			matrix2 = (i%2==0) ? matrix_b : matrix_c;

			Ctrl_Launch(ctrl, Mult, threads, CTRL_THREAD_NULL, matrix1, matrix_a, matrix2);
			Ctrl_HostTask(ctrl, Host_Compute, i, p_sum, p_res, matrix1, matrix_tmp);
		}

		Ctrl_GlobalSync(ctrl);
		exec_clock = omp_get_wtime() - exec_clock;

		Ctrl_Free(ctrl, matrix_a, matrix_b, matrix_c);
		hit_tileFree(matrix_tmp);

		Ctrl_Destroy(ctrl);
	}

	free(p_sum);
	free(p_res);

	main_clock = omp_get_wtime() - main_clock;

	printf("%lf, %lf\n", main_clock, exec_clock);
	return EXIT_SUCCESS;
}
